/*

		occult.c

   functions for creating and managing occultations of one
   fiber by another fiber.

*/


#include <stdio.h>
#include <strings.h>
/* #include <setjmp.h> */
/* #include <sys/time.h> */

#include "miscMath.h"
#include "plate.h"
#include "motor.h"
#include "kaboom.h"
#include "occult.h"

#ifdef ARCHIVE
#include "archive.h"
#endif


/***************** local prototypes *******************/

static void   createVictims( fibDB *perpetrator );
static void   releaseVictims( fibDB *perpetrator, struct anOccult *victims );
static void   liberate( fibDB *victim, fibDB *perpetrator );
static void   occultLink( fibDB *victim );
static void   occultRemove(  struct anOccult *oldOccult );
static struct anOccult  *dupEntry( struct anOccult *entry );
static void   appendEntry( struct anOccult *oldEntries, struct anOccult *newEntry );
static struct anOccult *occultDiff( struct anOccult *list1, struct anOccult *list2 );
static void   victimize( fibDB *victim, fibDB *perpetrator );
static void   occultFree( struct anOccult *oldOccult );
static void   occultFreeAll( struct anOccult *oldOccult );

/***************** end local prototypes *******************/

/*
   occultAddPending

   Create an occultation structure for the perpetrator,
   and link it in to the "pending" list for the specified fiber.
*/


void occultAddPending( fibDB *victim, fibDB *perpetrator )

{
   struct anOccult  *newEntry;

   /* create the new one */

   newEntry = occultAlloc(  victim );

   appendEntry( perpetrator->pendWeOccult, newEntry );
}



/*
   occultClearPending

   Clear the occultation pending list for the specified fiber
*/


void occultClearPending( fibDB *fiber )

{
   occultFree( fiber->pendWeOccult );

   fiber->pendWeOccult->next = NULL;
}



void recordNewOccultations( fibDB *perpetrator )

{
   struct anOccult *theFree;

   struct anOccult *tmpOccult;

   /* first clear the occultations that are no longer valid */

   theFree = occultDiff( perpetrator->weOccult, perpetrator->pendWeOccult );

   releaseVictims( perpetrator, theFree );

   occultFreeAll( theFree );

   /* now descend the "we occult" list and link in the new "occulted by"s */

   createVictims( perpetrator );

   /* the pending list become the master list */

   tmpOccult = perpetrator->weOccult;

   perpetrator->weOccult = perpetrator->pendWeOccult;

   perpetrator->pendWeOccult = tmpOccult;

   occultFree( perpetrator->pendWeOccult );

   perpetrator->pendWeOccult->next = NULL;

}


/*
	createVictims

   Go through the perpetrators "we occult" list and make victims of the
   indicated fibers, after first checking to see if they are already victims
   of this perpetrator.  Then flush the new "occulted by" list for the 
   victim to disk.
*/

void createVictims( fibDB *perpetrator )

{
   struct anOccult *potentialVictim;

   potentialVictim = perpetrator->pendWeOccult->next;
   
   /* loop through list */

   while( potentialVictim != NULL ){

      /* already a victim? */

      if( !alreadyVictimized( potentialVictim->fiber, perpetrator ) ){

	 victimize( potentialVictim->fiber, perpetrator );
      }

      /* find out who's next on the list */

      potentialVictim = potentialVictim->next;
   }
}


void releaseVictims( fibDB *perpetrator, struct anOccult *victims )

{
   /* loop through all the victims and release them */

   while( victims != NULL ){
      
      if( victims->fiber != NULL )
         liberate( victims->fiber, perpetrator );

      victims = victims->next;
   }
}


void liberate( fibDB *victim, fibDB *perpetrator )

{

   struct anOccult  *tmpOccult = victim->occultedBy;
   struct anOccult  *oldOppressor;
   
   while( tmpOccult != NULL ){
      
      if( tmpOccult->fiber == perpetrator ){
	 
	 oldOppressor = tmpOccult;

	 tmpOccult = tmpOccult->next;

	 occultRemove( oldOppressor );

#ifdef ARCHIVE
         archivePosition( victim, positionArchive );
#endif



	 return;
      }
      else{
	 tmpOccult = tmpOccult->next;
      }
   }
}



/*
   occultLink

   Take the specified occultation list and use it as a
   replacement for the victim's existing list.  
   Flush the new list to the archive entry for the victim.
   De-Alloc the memory used by the old list.

*/

void occultLink( fibDB *victim )

{
   struct anOccult  *oldOccult;

   if( victim->pendOccultBy != NULL ){

      oldOccult = victim->occultedBy;

      /* link in the new occultation stuff. */

      victim->occultedBy = victim->pendOccultBy;

      occultFree( oldOccult );

#ifdef ARCHIVE
      archivePosition( victim, positionArchive );
#endif

   }

   victim->pendOccultBy = NULL;

}



/*
   occultRemove

   Unlink and de-alloc the specified occultation from the 
   victim's list.  
*/

void occultRemove(  struct anOccult *oldOccult )

{
   /* don't allow removal of header record! */

   if( oldOccult->fiber == NULL )
      return;

   if( oldOccult->prev != NULL )
      oldOccult->prev->next = oldOccult->next;

   if( oldOccult->next != NULL )
      oldOccult->next->prev = oldOccult->prev;

   free( (char *)oldOccult );
}




#ifdef IGNORE

/*
   occultCopy

   Create a duplicate list for the victim and return a pointer
   to it.
*/

extern struct anOccult  *occultCopy( victim )

struct fiber_dbase *victim;

{
   struct anOccult  *newEntries = NULL;
   struct anOccult  *nextItem;

   nextItem = victim->occultedBy;

   while( nextItem != NULL ){

      appendEntry( newEntries, dupEntry( nextItem ) );

      nextItem = nextItem->next;
   }

   return( newEntries );
}


#endif


extern struct anOccult  *dupEntry( struct anOccult *entry )

{

   struct anOccult  *newEntry;

   newEntry = (struct anOccult *)malloc(
       (unsigned)sizeof( struct anOccult ));

   newEntry->next = NULL;

   newEntry->fiber = entry->fiber;

   return( newEntry );
}



void appendEntry( struct anOccult *oldEntries, struct anOccult *newEntry )

{

   struct anOccult  *tailPtr;

   /* start at the beginning of the list */

   tailPtr = oldEntries;

   /* find the end of the list */

   while( tailPtr->next != NULL ){
      tailPtr = tailPtr->next;
   }

   /* now link in the new entry */

   /* set the forward link */

   tailPtr->next = newEntry;

   /* set the backward link */

   tailPtr->next->prev = tailPtr;
}



/*
	occultDiff

   Create an occultation list that consists of the occultations in
   list1 that are not present in list2.  This function is typically
   used to generate a list of fibers that are no longer occulted
*/


struct anOccult *occultDiff( struct anOccult *list1, struct anOccult *list2 )

{
   struct anOccult  *tmpList;
   struct anOccult  *outList = occultAlloc( (struct fiber_dbase *)NULL );
   int bingo;

   while( list1 != NULL ){

      tmpList = list2;
      bingo = FALSE;

      while( tmpList != NULL ){

	 if( list1->fiber == tmpList->fiber ){
	    bingo = TRUE;
	    break;
	 }
	 else{
	    tmpList = tmpList->next;
	 }
      }

      if( !bingo ){
	 appendEntry( outList, dupEntry( list1 ) );
      }

      list1 = list1->next;
   }
   
   return( outList );
}



/*
	alreadyVictimized

   See if the potential victim is already a victim of the
   perpetrator.  Return true or false.
*/

int alreadyVictimized( fibDB *potentialVictim, fibDB *perpetrator )

{
   struct anOccult *tmpList;

   tmpList = potentialVictim->occultedBy;

   while( tmpList != NULL ){

      if( tmpList->fiber == perpetrator ){
	 return( TRUE );
      }
      else{
	 tmpList = tmpList->next;
      }
   }

   return( FALSE );
}



/*
   victimize

   Create an occultation structure containing the perpetrator,
   and link it in to the existing occultations of the victim.
*/


void victimize( fibDB *victim, fibDB *perpetrator )

{
   /* create the new one */

   appendEntry( (victim->occultedBy), occultAlloc(  perpetrator ) );

#ifdef ARCHIVE
      archivePosition( victim, positionArchive );
#endif
}


/*
   occultList

   Create a text representation of the occultations for the
   victim.  return a pointer to this string
*/

char *occultList( fibDB *victim )

{
   static char theList[128];

   char tmpString[32];

   struct anOccult  *nextItem;

   *theList = '\0';
   nextItem = victim->occultedBy;

   while( nextItem != NULL ){

      if( nextItem->fiber == NULL ){

	 /* skip the header record */

	 /* fputs("skipping header\n", stderr );*/

         nextItem = nextItem->next;
	 continue;
      }

      sprintf( tmpString, "%3d:", nextItem->fiber->fiber_id ); 

      strcat( theList, tmpString );

      nextItem = nextItem->next;
   }

   if( *theList == '\0' )
      strcpy( theList, "no_occult" );
   
   return( theList );
}


extern struct anOccult  *occListToStruct( char *rawList )

{
   char ourList[64];
   char *list;
   char *nextToken;
   int id;
   struct anOccult *occultations = occultAlloc( NULL );

   strcpy( ourList, rawList );
   list = ourList;

   while( ( nextToken = (char *)strtok( list, ":" ) ) != NULL ){

      if( !isdigit( nextToken[0] ) )
	 return( occultations );
	 
      id = atoi( nextToken );

      /* fprintf( stderr, "Occult parsed fiber #%d \n", id ); */


      appendEntry( occultations, occultAlloc( &FiBase[id] ) );

      list = NULL;
   }

   return( occultations );
}


/*
	applyOccultationList

   Take the list of occultations that was read in from
   a locations file and link it into the victim's 
   occultedBy list.  Next, take the list and make
   sure there is a matching entry in the weOccult
   list of the oppressing fibers.
*/

void applyOccultationList( fibDB *fib, struct anOccult *theList )

{
   if( theList == NULL ){

      /* don't bother with empty lists */

      occultFreeAll( theList );

   }
   else{

      occultFreeAll( fib->occultedBy );

      fib->occultedBy = theList;

      while( theList != NULL ){

	 if( theList->fiber != NULL )
	    appendEntry( theList->fiber->weOccult, occultAlloc( fib ) );
	 
	 theList = theList->next;
      }
   }

}


/*
	occultMessage

   print a pretty message saying who is occulting
   the victim.
*/

void occultMessage( fibDB *victim )

{
   struct anOccult *next = victim->occultedBy->next;

#ifdef LOG
   log_fprintf( stderr, "Fiber #%d is occulted by fiber(s): ", 
#else
   fprintf( stderr, "Fiber #%d is occulted by fiber(s): ", 
#endif
      victim->fiber_id );
   
   while( next != NULL ){

      if( next->next != NULL )

#ifdef LOG
         log_fprintf( stderr, "%d, ", next->fiber->fiber_id );
#else
         fprintf( stderr, "%d, ", next->fiber->fiber_id );
#endif
      else

#ifdef LOG
         log_fprintf( stderr, "%d.\n", next->fiber->fiber_id );
#else
         fprintf( stderr, "%d.\n", next->fiber->fiber_id );
#endif

      next = next->next;
   }
}



int isOcculted( fibDB *victim )

{
   return( victim->occultedBy->next != NULL );
}




/*
   occultAlloc

   Create an occultation structure for the perpetrator,
   and return a pointer to it.
*/


struct anOccult *occultAlloc(  fibDB *perpetrator )

{
   struct anOccult  *newEntry;

   /* create the new one */

   newEntry = (struct anOccult *)malloc(
       (unsigned)sizeof( struct anOccult ));

   newEntry->next = NULL;
   newEntry->prev = NULL;

   newEntry->fiber = perpetrator;

   return ( newEntry );
}



/*
   occultFree

   release the the memory used by the specified occultationList
   Do not release the header record.

*/

void occultFree( struct anOccult *oldOccult )

{

   struct anOccult  *target;
   struct anOccult  *next;

   target = oldOccult;

   while( target != NULL ){

      if( target->fiber == NULL ){
         target = target->next;
	 continue;
      }

      next = target->next;
      free( target );
      target = next;
   }
   
}



/*
   occultFreeAll

   release the the memory used by the specified occultationList
   *including" the header record

*/

void occultFreeAll( struct anOccult *oldOccult )

{

   struct anOccult  *target;
   struct anOccult  *next;

   target = oldOccult;

   while( target != NULL ){

      next = target->next;
      free( target );
      target = next;
   }
   
}
